/* ------------------------------------------------------------------------
@NAME       : 
@DESCRIPTION: 
@GLOBALS    : 
@CREATED    : 
@MODIFIED   : 
@VERSION    : $Id: file_header.c 6328 2008-10-08 21:07:28Z ambs $
@COPYRIGHT  : Copyright (c) 1996-99 by Gregory P. Ward.  All rights reserved.

              This file is part of the btparse library.  This library is
              free software; you can redistribute it and/or modify it under
              the terms of the GNU Library General Public License as
              published by the Free Software Foundation; either version 2
              of the License, or (at your option) any later version.
-------------------------------------------------------------------------- */

#include "bt_config.h"


#include "my_dmalloc.h"
